""" 
flash.py

A port of Ruby on Rails' Flash.
(see ticket #91, http://subway.python-hosting.com/ticket/91)

Flashes are simple dict-like objects primarily intended to be used in your
Subway application to store status messages. 

Best explained in the very Rails Action Pack docs [1]: 
    ''The flash provides a way to pass temporary objects between actions.
    Anything you place in the flash will be exposed to the very next action
    and then cleared out. This is a great way of doing notices and alerts,
    such as a create action that sets flash["notice"] = "Successfully
    created" before redirecting to a display action that can then expose
    the flash to its template.''
[1] http://api.rubyonrails.com/classes/ActionController/Flash.html

How to use it
-------------

Say you have an hanlder which saves an entry to your database. A common
idiom is to redirect to another handler after the creation has occurred.
Assuming that 'flash' is a global name referencing a Flash object, in the
controller class we have do the following:
    
    @expose
    def save_thing(thing_name):
        
        # ... saving code ...
        
        flash['notice'] = "Your thing %s has been saved" % thing_name
        return redirect("/list_things")
    
    @expose
    def list_things():
        things = model.Thing.select()
        return view.list_things()
    
Then, in your template:

    $widget.notice
    
    #for $thing in $things
        ## display the thing...
    #end for
     
Now, if 'flash' were a normal dictionary, you'd have to be careful
to delete the 'notice' item after it's used. Not with a Flash object.
After the `list_things` action, the 'notice' key will be automatically
deleted.
""" 

from cherrypy.lib.filter import basefilter
import cherrypy

FLASH_KEY = '_flash'

class FlashFilter(basefilter.BaseFilter):
        
    #@@ is "flash_session_key" param useful at all?
    def __init__(self, flash_session_key='_flash'):
        #self.flash_session_key = flash_session_key
        pass

    def beforeMain(self):
        # skip static content
        if (not cherrypy.config.get('sessionFilter.on', False)) or \
                cherrypy.config.get('staticFilter.on', False):
            return
        
        flash = cherrypy.session.get(FLASH_KEY, None)
        if flash is None:
            cherrypy.session[FLASH_KEY] = Flash()

class Flash(dict):
    def __getitem__(self, key):
        if key in self:             
            # grab value and then throw away the key
            value = self.get(key)
            del self[key]
            return value
        else:
            raise AttributeError("No key %r" % key)

