import sys
import os
import glob

import cherrypy
import cherrypy._cpwsgi

import subway

import thread

import namespaces

importing = None # lazy import

def init_app(home_dir):
    """
    Does all the work that is common between loading apps and loading 
    """
    # set up the staticFilter first so the user can override it
    static_filter_settings = {
        "staticFilter.on" : True,
        "staticFilter.dir" : os.path.join(home_dir, "public")
    }
    
    # custom input filters
    input_filters = {
        'auth' : 'AuthFilter',
        'flash': 'FlashFilter',
    }
    
    # custom output filters
    output_filters = {        
    }
    
    # set up global settings
    global_settings = {
        "sessionFilter.on" : True,
        "logDebugInfoFilter.on" : False,
        
        "server.inputFiltersDict": input_filters,
        "server.outputFiltersDict": output_filters,
            
        "server.filtersRoot": [os.path.dirname(subway.__file__)],  
    }

    cherrypy.config.update({
        "/public": static_filter_settings,
        "global": global_settings
    })
    
    # load the config first BEFORE the framework imports
    config_file = os.path.join(home_dir, "application.conf")
    cherrypy.config.update(file = config_file)
    subway.easy_filter_config()

    # this read config information, so we have to
    # import them after loading the config
    global importing
    import importing
    
    # now, add the lib directory to the pythonpath
    lib_dir = os.path.join(home_dir, "lib")
    sys.path.insert(0, lib_dir)

    # add web widgets to the default namespace
    widget_dir = os.path.join(home_dir, "widget")
     # deprecated, use "widget"instead
    namespaces.DEFAULT_VIEW_NAMESPACE["widgets"] = subway.ViewManager(widget_dir)
    namespaces.DEFAULT_VIEW_NAMESPACE["widget"] = subway.ViewManager(widget_dir)  

    # load the model definitions
    subway.sql_connect(-1)
    model_dir = os.path.join(home_dir, "model")
    model_files = glob.glob(os.path.join(model_dir, "*.py"))
    model = importing.import_models(model_files)
    subway.model = model # make it accessible to the outside

class Empty:
    pass

def load_controller_dir(parent_obj, current_view_dir, controller_dir):
    name = os.path.basename(controller_dir)
    if not hasattr(parent_obj, name):
        setattr(parent_obj, name, Empty())
    current = getattr(parent_obj, name)
    for filename in os.listdir(controller_dir):
        controller_file = os.path.join(controller_dir, filename)
        if controller_file.endswith(".py") and os.path.isfile(controller_file):
            controller_name = os.path.splitext(os.path.basename(controller_file))[0]
            if controller_name == "default":
                if hasattr(parent_obj, name) and not isinstance(getattr(parent_obj, name), Empty):
                    cherrypy.log("WARNING: name conflict for %r" % name)
                else:
                    setattr(parent_obj, name, importing.import_controller(controller_file,
                                                                          current_view_dir,
                                                                          subway.model))
                    current = getattr(parent_obj, name)
            else:
                if hasattr(current, controller_name) and not isinstance(getattr(current, controller_name), Empty):
                    cherrypy.log("WARNING: name conflict for %r" % controller_name)
                else:
                    setattr(current,
                            controller_name,
                            importing.import_controller(controller_file,
                                                        os.path.join(current_view_dir, controller_name),
                                                        subway.model))
        elif os.path.isdir(controller_file):
            load_controller_dir(current, os.path.join(current_view_dir, filename), controller_file)
        
def load_webapp(home_dir):
    cherrypy.log("Loading Subway web application from %s." % home_dir)

    # now load the root controller
    controller_dir = os.path.join(home_dir, "controller")
    view_dir = os.path.join(home_dir, "view")
    empty = Empty()
    load_controller_dir(empty, view_dir, controller_dir)
    cherrypy.root = empty.controller
    # at this point the app is loaded! hurray!

def load_webapp_server(home_dir):
    load_webapp(home_dir)
    cherrypy.server.start()

def load_webapp_wsgi(home_dir):
    load_webapp(home_dir)
    cherrypy.server.start(initOnly = True)
    return cherrypy._cpwsgi.wsgiApp

def load_daemons(home_dir, daemons=None, main_thread=False):
    daemon_path = os.path.join(home_dir, "daemons")
    if daemons is None:
        # list all daemons in the directory
        daemons = [os.path.splitext(os.path.basename(filename)) for filename in glob.glob(os.path.join(daemon_path, "*.py"))]
    cherrypy.log("Loading Subway daemons %s from %s." % (",".join(daemons), home_dir))
    if main_thread:
        for daemon in daemons[:-1]:
            thread.start_new_thread(importing.import_daemon(os.path.join(daemon_path, daemon + ".py"), subway.model), ())
        # run the last daemon in the main thread
        importing.import_daemon(os.path.join(daemon_path, daemons[-1] + ".py"), subway.model)()
    else:
        for daemon in daemons:
            thread.start_new_thread(importing.import_daemon(os.path.join(daemon_path, daemon + ".py"), subway.model), ())
    # at this point all daemons are loaded 
    