# view.py - easily accessible precompiled cheetah templates
import os
import imp
import sys

namespaces = None # lazy import

from Cheetah import Compiler, NameMapper, Template

import cherrypy

import flash, forms

from formencode import htmlfill, htmlfill_schemabuilder

import re

RE_EXTENDS = re.compile(r"#extends\s+(.+)")
VIEW_HEADER = """import os
from subway import view as subway_view
baseviewdir,viewname = os.path.split(os.path.dirname(os.path.abspath(__file__)))
view = getattr(subway_view.ViewManager(baseviewdir), viewname)
"""

def get_cheetah_template(file):
    """
    Returns subclass of Template
    file - source file to compile

    this method cheetah compiles the source file if it is newer than the compiled version    
    """
    tmpldir, filename = os.path.split(file)
    pathnoext, ext = os.path.splitext(file)
    tname = os.path.basename(pathnoext) + "_tmpl"
    compiledfile = os.path.join(tmpldir, tname + ".py")
    if not os.path.exists(compiledfile) or os.stat(file).st_mtime > os.stat(compiledfile).st_mtime:
        # compiled file either does not exist, or is older than the source.
        # recompile the template
        source = open(file, "r").read()
        # this next chunk figures out all of the classes
        # this template #extends, and puts it into extendlist
        extendlist = []
        def _sub(match):
            if match.group(1).startswith("view"):
                # if extending a view-prefixed class, munge the name
                # because view-prefixed base classes need to use magic
                e = match.group(1)[match.group(1).find(".")+1:].replace(".", "_dot_")
                extendlist.append(e)
                return "#extends " + e
            else:
                return match.group(0)
        source = RE_EXTENDS.sub(_sub, source)
        # now, compile the file
        code = str(Compiler.Compiler(source=source, moduleName=tname, mainClassName=tname))
        
        # replace cheetah-generated view imports to subway view lookups
        for extend in extendlist:
            code = code.replace("from %s import %s" % (extend, extend), "%s = view.%s.template_class" % (extend, extend.replace("_dot_",".")))
        # write out the compiled file including boilerplate header
        open(compiledfile, 'w').write(VIEW_HEADER + code)
    # import the template
    fp, pathname, description = imp.find_module(tname, [tmpldir])
    try:
        a = imp.load_module(tname, fp, pathname, description)
    finally: # Since we may exit via an exception, close fp explicitly.
        if fp:
            fp.close()
    return getattr(a, tname) # return the class extracted from the module

class CallableTemplate(object):
    '''
    This is the base class of a Cheetah template
    you can access via the following API:
    mytemplate.template_class - returns the Cheetah template class, or equivalent
                                (see StringTemplate source for an exception)
    mytemplate._parent - return the parent CallableTemplate
    mytemplate.default_namespace - default namespace for the template
    mytemplate.form - return the FormEncode schema corresponding to this view
    mytemplate([], x=y) - return the result of executing this template. first argument is searchList, and it adds its
                          kwargs to the searchList
    mytemplate[x] = y - add a key to the default_namespace
    '''
    def __init__(self, default_namespace=None, parent=None):
        global namespaces
        if namespaces is None:
            from framework import namespaces
        if default_namespace:
            self.default_namespace = default_namespace
        else:
            self.default_namespace = namespaces.get_view_namespace()
        self._parent = parent
        self._form = None
    def __call__(self, *searchList, **firstItem):
        sl = [firstItem]
        sl.extend(searchList)
        sl.append(sys._getframe(firstItem.get("_frame", 1)).f_locals)
        sl.append(self.default_namespace)
        content = self.template_class(searchList=sl).respond()
        forms.form_defaults.update(cherrypy.request.paramMap)
        if "_resume" in forms.form_defaults:
            del forms.form_defaults["_resume"] # we don't want to carry over the resume id
        if forms.form_errors or forms.form_defaults:
            errors = forms.form_errors
            if not errors:
                errors = None
            listener = htmlfill_schemabuilder.SchemaBuilder()
            p = htmlfill.FillingParser(forms.form_defaults, listener=listener, errors=errors)
            p.feed(content)
            self._form = listener.schema()
            content = p.text()
            p.close()
        return content
    def _get_form(self):
        raise NotImplementedError, "You must implement a form property or attribute."
    form = property(_get_form)
    def _get_template_class(self):
        raise NotImplementedError, "You must implement a template_class property or attribute."
    template_class = property(_get_template_class)
    def __getitem__(self, key):
        return self.default_namespace[key]
    def __setitem__(self, key, value):
        self.default_namespace[key] = value

class FileCallableTemplate(CallableTemplate):
    '''
    A CallableTemplate which uses a template file on-disk.
    Construct it by providing a template filename.
    '''
    def __init__(self, filename, *args, **kwargs):
        CallableTemplate.__init__(self, *args, **kwargs)
        self.filename = filename
    def _get_form(self):
        if not self._form:
            listener = htmlfill_schemabuilder.SchemaBuilder()
            p = htmlfill.FillingParser({}, listener=listener)
            p.feed(open(self.filename, "r").read())
            p.close()
            self._form = listener.schema()
        return self._form
    form = property(_get_form)
    def _get_template_class(self):
        return get_cheetah_template(self.filename)
    template_class = property(_get_template_class)

class StringTemplate(CallableTemplate):
    '''
    A template that uses strings.
    Construct with a string containing the source code of the template.
    '''
    def __init__(self, s, *args, **kwargs):
        CallableTemplate.__init__(self, *args, **kwargs)
        self.s = s
        # A REMINDER:
        # if you plan on precompiling me in the future, precompile
        # ONLY when it is called the first time (not in the constructor)
    def _get_form(self):
        if not self._form:
            listener = htmlfill_schemabuilder.SchemaBuilder()
            p = htmlfill.FillingParser({}, listener=listener)
            p.feed(self.s)
            p.close()
            self._form = listener.schema()
        return self._form
    form = property(_get_form)
    def _get_template_class(self):
        return self._render
    template_class = property(_get_template_class)
    def _render(self, searchList):
        return Template.Template(self.s, searchList)

class ViewManager(object):
    '''
    This class represents directories you can use to manage views.
    Usage:
    mgr = ViewManager(path_to_dir, optional_list_of_template_extensions, optional_default_namespace)
    mgr._parent yields parent VIewManager, or None
    Subdirectories and template files are both accessed as mgr.name. Return values are
    either another ViewManager instance, or a FileCallableTemplate instance.
    '''
    def __init__(self, path=None, exts=[".htm",".html",".tmpl",".xml",".txt"], default_namespace=None, parent=None):
        global namespaces
        if namespaces is None:
            from framework import namespaces
        self.path = path
        self.exts = exts
        
        if default_namespace:
            self.default_namespace = default_namespace
        else:
            self.default_namespace = namespaces.get_view_namespace()
            
        self._parent = parent
    def __getattribute__(self, key):
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            if not self.path:
                self.path = os.getcwd()
            newpath = os.path.join(self.path, key)
            rv = None
            for ext in self.exts:
                if os.path.isfile(newpath + ext):
                    rv = FileCallableTemplate(newpath + ext, self.default_namespace, self)
            if os.path.isdir(newpath):
                rv = ViewManager(newpath, self.exts, self.default_namespace, self)
            if not rv:
                if not self._parent:
                    raise AttributeError, key
                else:
                    rv = getattr(self._parent, key) # acquisition
            setattr(self, key, rv)
            return rv
    def __getitem__(self, key):
        return self.default_namespace[key]
    def __setitem__(self, key, value):
        self.default_namespace[key] = value
        
def docstring_view(m):
    """
    When you apply this decorator, the method will now have a .view method
    corresponding to a Cheetah template in its docstring. This allows you
    to do stuff like:

    @cherrypy.expose
    @subway.docstring_view
    def mymethod(self):
        '''
        <html><body>Hello, $name !</body></html>
        '''
        name = "Subway"
        return self.mymethod.view()
        
    This decorator is implied in Subway's subway.expose decorator.
    """
    if hasattr(m, "__doc__"):
        m.view = StringTemplate(m.__doc__)
    return m
